/*
 * $Id: httpd.c,v 1.5 2003/03/07 02:34:06 jasta Exp $
 *
 * Copyright (C) 2001-2003 giFT project (gift.sourceforge.net)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include "gift.h"

#include "file.h"
#include "connection.h"
#include "fdbuf.h"

#include "httpd.h"

/*****************************************************************************/

static TCPConn *httpd = NULL;

/*****************************************************************************/

static char *get_fortune (char *def)
{
	FILE   *p;
	String *fortune;
	char   *line = NULL;

	if (!(p = popen ("/bin/sh -c fortune", "r")))
		return STRDUP(def);

	if (!(fortune = string_new (NULL, 0, 0, TRUE)))
		return STRDUP(def);

	while (file_read_line (p, &line))
		string_appendf (fortune, "%s", line);

	pclose (p);

	return string_free_keep (fortune);
}

static int send_fortune (TCPConn *c)
{
	char *msg;
	char *fortune;

	fortune = get_fortune ("He said nothing.");
	assert (fortune != NULL);

	msg = stringf ("HTTP/1.0 200 OK\r\n"
	               "Content-Type: text/html\r\n"
	               "\r\n"
	               "<h1>And a wise man once said...</h1><pre>%s</pre>\n",
	               fortune);

	free (fortune);

	return tcp_writestr (c, msg);
}

/*****************************************************************************/

static int check_sentinel (char *data, size_t len)
{
	int cnt;

	assert (len > 0);
	len--;

	for (cnt = 0; len > 0 && cnt < 2; cnt++)
	{
		if (data[len--] != '\n')
			break;

		/* treat CRLF as LF */
		if (data[len] == '\r')
			len--;
	}

	return (cnt == 2);
}

static void handle_http (int fd, input_id id, TCPConn *c)
{
	FDBuf *buf;
	char  *data;
	size_t data_len = 0;
	int    n;

	buf = tcp_readbuf (c);

	if ((n = fdbuf_delim (buf, "\n")) < 0)
	{
		tcp_writestr (c, "Booya!\n");
		tcp_close (c);
		return;
	}

	if (n > 0)
		return;

	data = (char *)fdbuf_data (buf, &data_len);

	/* look for two trailing \n's optionally preceeded by \r characters */
	if (!check_sentinel (data, data_len))
		return;

	fdbuf_release (buf);

	TRACE_MEM (data, data_len);

	send_fortune (c);
	tcp_close (c);
}

static void handle_incoming (int fd, input_id id, TCPConn *c)
{
	TCPConn *accept;

	if (!(accept = tcp_accept (c, FALSE)))
		return;

	input_add (accept->fd, accept, INPUT_READ,
	           (InputCallback)handle_http, TRUE);
}

/*****************************************************************************/

int httpd_start (in_port_t port)
{
	if (!(httpd = tcp_bind (port, FALSE)))
		return FALSE;

	input_add (httpd->fd, httpd, INPUT_READ,
	           (InputCallback)handle_incoming, FALSE);

	return TRUE;
}

void httpd_stop (void)
{
	tcp_close_null (&httpd);
}

/*****************************************************************************/

#if 0
int main (int argc, char **argv)
{
	in_port_t port = 8080;

	log_init (GLOG_STDOUT, "httpd", 0, 0, NULL);
	event_init ();

	if (!httpd_start (port))
	{
		GIFT_ERROR (("unable to start http server on port %hu: %s",
		             port, GIFT_NETERROR()));
		return 1;
	}

	event_loop ();

	return 0;
}
#endif
